import json
import numpy as np
import pytest

from src.ersilia_pack_utils.core import write_out_bin, _normalize_dtype, clip


def _read_bin_payload(path):
    with open(path, "rb") as f:
        meta_line = f.readline()
        meta = json.loads(meta_line.decode("utf-8"))
        dt = np.dtype(meta["dtype"])
        shape = tuple(meta["shape"])
        arr = np.frombuffer(f.read(), dtype=dt).reshape(shape)
    return meta, arr


def test_normalize_dtype_accepts_int32_float32():
    assert _normalize_dtype("int32") == np.dtype(np.int32)
    assert _normalize_dtype("float32") == np.dtype(np.float32)
    assert _normalize_dtype(np.int32) == np.dtype(np.int32)
    assert _normalize_dtype(np.float32) == np.dtype(np.float32)


def test_write_out_bin_rejects_other_dtypes(tmp_path):
    out = tmp_path / "x.bin"
    results = [[1.0, 2.0]]
    header = ["a", "b"]
    with pytest.raises(ValueError):
        write_out_bin(results, header, str(out), dtype="float64")
    with pytest.raises(ValueError):
        write_out_bin(results, header, str(out), dtype=np.int64)


def test_float32_clips_and_handles_infs_and_nans(tmp_path):
    out = tmp_path / "f.bin"
    header = ["x", "y"]
    f32 = np.finfo(np.float32)

    big = np.float64(f32.max) * 2.0
    neg_big = -np.float64(f32.max) * 2.0

    results = [
        [0.0, big],
        [neg_big, 1.0],
        [np.inf, -np.inf],
        [np.nan, 123.0],
    ]

    write_out_bin(results, header, str(out), dtype=np.float32)
    meta, arr = _read_bin_payload(out)

    assert meta["dtype"] == "float32"
    assert meta["shape"] == [4, 2]
    assert meta["columns"] == header

    assert arr.dtype == np.float32
    assert arr[0, 0] == np.float32(0.0)
    assert arr[0, 1] == np.float32(f32.max)
    assert arr[1, 0] == np.float32(f32.min)
    assert arr[1, 1] == np.float32(1.0)
    assert arr[2, 0] == np.float32(f32.max)
    assert arr[2, 1] == np.float32(f32.min)
    assert np.isnan(arr[3, 0])
    assert arr[3, 1] == np.float32(123.0)


def test_int32_casts_directly(tmp_path):
    out = tmp_path / "i.bin"
    header = ["x", "y"]

    results = [
        [1.9, -2.2],
        [3.0, 4.0],
    ]

    write_out_bin(results, header, str(out), dtype=np.int32)
    meta, arr = _read_bin_payload(out)

    assert meta["dtype"] == "int32"
    assert arr.dtype == np.int32
    assert arr.tolist() == [[1, -2], [3, 4]]


def test_header_length_mismatch_raises(tmp_path):
    out = tmp_path / "h.bin"
    results = [[1.0, 2.0]]
    header = ["only_one_col"]
    with pytest.raises(ValueError):
        write_out_bin(results, header, str(out), dtype=np.float32)


def test_1d_results_are_reshaped_to_column(tmp_path):
    out = tmp_path / "c.bin"
    header = ["x"]
    results = [1.0, 2.0, 3.0]

    write_out_bin(results, header, str(out), dtype=np.float32)
    meta, arr = _read_bin_payload(out)

    assert meta["shape"] == [3, 1]
    assert arr.shape == (3, 1)
    assert np.all(arr[:, 0] == np.array([1.0, 2.0, 3.0], dtype=np.float32))


def test_coerce_results_float32_clips_without_write():
    f32 = np.finfo(np.float32)
    big = np.float64(f32.max) * 10.0
    x = np.array([big, -big, np.inf, -np.inf, np.nan], dtype=np.float64)

    out = clip(x, np.dtype(np.float32))

    assert out.dtype == np.float32
    assert out[0] == np.float32(f32.max)
    assert out[1] == np.float32(f32.min)
    assert out[2] == np.float32(f32.max)
    assert out[3] == np.float32(f32.min)
    assert np.isnan(out[4])



import csv
import json
import struct
import numpy as np
from typing import List, Tuple, Optional, Any

Smiles = List[str]
Header = List[str]

_ALLOWED_BIN_DTYPES = {
    np.dtype(np.int32),
    np.dtype(np.float32),
}

_DTYPE_ALIASES = {
    "int32": np.dtype(np.int32),
    "float32": np.dtype(np.float32),
}

_F32 = np.finfo(np.float32)


def _normalize_dtype(dtype: Any) -> np.dtype:
    if dtype is None:
        raise ValueError("dtype must be specified (only 'int32' or 'float32').")

    if isinstance(dtype, str):
        key = dtype.strip().lower()
        if key in _DTYPE_ALIASES:
            return _DTYPE_ALIASES[key]
        return np.dtype(key)

    return np.dtype(dtype)


def _clip(results):
    x = np.asarray(results, dtype=np.float64)
    try:
        np.nan_to_num(x, copy=False, nan=np.nan, posinf=_F32.max, neginf=_F32.min)
        np.clip(x, _F32.min, _F32.max, out=x)
        return x.astype(np.float32)
    except TypeError:
        nan_mask = np.isnan(x)
        np.nan_to_num(x, copy=False, posinf=_F32.max, neginf=_F32.min)
        np.clip(x, _F32.min, _F32.max, out=x)
        x[nan_mask] = np.nan
        return x.astype(np.float32)

def clip(results, dt: np.dtype) -> np.ndarray:
    if dt == np.dtype(np.int32):
        return np.asarray(results, dtype=np.int32)
    if dt == np.dtype(np.float32):
        return _clip(results)
    raise ValueError(f"Unsupported dtype {dt!r}.")

def read_smiles_csv(in_file: str) -> Tuple[Header, Smiles]:
    with open(in_file, "r", newline="") as f:
        reader = csv.reader(f)
        cols = next(reader)
        data = [r[0] for r in reader]
    return cols, data


def read_smiles_bin(in_file: str) -> Tuple[Header, Smiles]:
    with open(in_file, "rb") as f:
        data = f.read()

    mv = memoryview(data)
    nl = mv.tobytes().find(b"\n")
    meta = json.loads(mv[:nl].tobytes().decode("utf-8"))
    cols = meta.get("columns", [])
    count = meta.get("count", 0)

    smiles_list = [None] * count
    offset = nl + 1
    for i in range(count):
        (length,) = struct.unpack_from(">I", mv, offset)
        offset += 4
        smiles_list[i] = mv[offset: offset + length].tobytes().decode("utf-8")
        offset += length

    return cols, smiles_list


def read_smiles(in_file: str) -> Tuple[Header, Smiles]:
    if in_file.endswith(".bin"):
        return read_smiles_bin(in_file)
    return read_smiles_csv(in_file)


def write_out_csv(results, header, file: str) -> None:
    with open(file, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(header)
        writer.writerows(results)


def write_out_bin(results, header, file: str, dtype: Optional[Any] = None) -> None:
    dt = _normalize_dtype(dtype)

    if dt not in _ALLOWED_BIN_DTYPES:
        allowed = ", ".join(sorted(d.name for d in _ALLOWED_BIN_DTYPES))
        raise ValueError(f"Unsupported dtype {dt!r}. Allowed: {allowed}")

    arr = clip(results, dt)
    if arr.ndim == 1:
        arr = arr.reshape(-1, 1)

    if len(header) != arr.shape[1]:
        raise ValueError(f"Header length ({len(header)}) != number of columns ({arr.shape[1]}).")

    meta = {"columns": list(header), "shape": list(arr.shape), "dtype": dt.name}
    meta_bytes = (json.dumps(meta) + "\n").encode("utf-8")

    with open(file, "wb") as f:
        f.write(meta_bytes)
        f.truncate(len(meta_bytes) + arr.nbytes)

    m = np.memmap(file, dtype=dt, mode="r+", offset=len(meta_bytes), shape=arr.shape)
    m[:] = arr
    m.flush()


def write_out(results, header, file: str, dtype: Optional[Any] = np.float32) -> None:
    if file.endswith(".bin"):
        write_out_bin(results, header, file, dtype)
    elif file.endswith(".csv"):
        write_out_csv(results, header, file)
    else:
        raise ValueError(f"Unsupported extension for {file!r}")
